/*
libDataheapPlugin.so,
a library to count dataheap events in vampir trace
Copyright (C) 2010 TU Dresden, ZIH

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

 */

#include <vampirtrace/vt_plugin_cntr.h>
#include <expec.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>

#define MAX_EXPEC_COUNTER 20

typedef struct expec_counter_t {
    uint64_t expec_id;
    union{
      double dbl;
      uint64_t uint64;
    } last_value;
} expec_counter;

static expec_counter expec_counters[MAX_EXPEC_COUNTER];
static int used_expec_counters = 0;
static int init_libexpec_failed = 0;

static pthread_key_t pthreadKey;

void vt_newdata_callback(uint64_t dataSourceID, uint64_t timestamp, double value)
{
    int counter_pos;
    int found = 0;

    /* get the local counter structure for this counter */
    for (counter_pos = 0; counter_pos < used_expec_counters; counter_pos++)
    {
        if (expec_counters[counter_pos].expec_id == dataSourceID)
        {
          found=1;
          expec_counters[counter_pos].last_value.dbl=value;
          break;
        }
    }
    if (!found)
    {
        fprintf(stderr,"new data callback for unknown counter id %llu", dataSourceID);
        return;
    }
}



void init_libexpec_thread()
{
    uint8_t *is_libexpec_thread_flag;

    /* create thread-specific data key for all threads */
    if (pthread_key_create(&pthreadKey, NULL) != 0)
      init_libexpec_failed=1;


    is_libexpec_thread_flag = (uint8_t*)malloc(sizeof(uint8_t));
    if (is_libexpec_thread_flag == NULL)
      init_libexpec_failed=1;

    *is_libexpec_thread_flag = 1;

    /* put master thread-ID to thread-specific data */
    if (pthread_setspecific(pthreadKey, is_libexpec_thread_flag) != 0) {
      init_libexpec_failed=1;
    }
}

int init()
{
    char*    ext_metric_server;
    char*    colon_pos;
    uint32_t port;

    /* get server */
    ext_metric_server = getenv("VT_PLUGIN_CNTR_DATAHEAP_SERVER");
    if (ext_metric_server == NULL) {
        return 1;
    }
    colon_pos = strchr(ext_metric_server, ':');
    if (colon_pos == NULL) {
        /* Note by tilsche: Should we maybe set a default port here? */
        fprintf(stderr,
                "FATAL: ext counter server definition: could not find the ':'");
        return 1;
    }
    ext_metric_server = strdup(ext_metric_server);

    /* extract port number from "server:port" string */
    colon_pos = strchr(ext_metric_server, ':');
    *colon_pos = 0;
    colon_pos++;
    port = atoi(colon_pos);

    /* initialize connection */
    expec_initialize(ext_metric_server, port, init_libexpec_thread, 0);

    if (init_libexpec_failed) return 1;

    /* initialize callback */
    expec_registerNewDataCallback(vt_newdata_callback);

    free(ext_metric_server);

        return 0;
}

vt_plugin_cntr_metric_info * get_event_info(char * event_name)
{
    vt_plugin_cntr_metric_info * return_values;
    /* Wildcards are not supported yet, so we need two entries */
    return_values = malloc(2 * sizeof(vt_plugin_cntr_metric_info));

    /* if the description is null it should be considered the end */
    return_values[0].name = strdup(event_name);
    return_values[0].unit = strdup(expec_unit(event_name));
    return_values[0].cntr_property = VT_PLUGIN_CNTR_ABS
            | VT_PLUGIN_CNTR_DOUBLE | VT_PLUGIN_CNTR_LAST;
    /* Last element empty */
    return_values[1].name = NULL;
    return return_values;
}

void fini()
{
    expec_finalize();
}

int add_counter(char * event_name)
{
    uint64_t expec_id;
    uint64_t local_id;

    if (used_expec_counters == MAX_EXPEC_COUNTER)
    {
        fprintf(stderr,"maximum count of %d external counters reached", MAX_EXPEC_COUNTER);
        return -1;
    }

    expec_id = expec_subscribe(event_name, 0);
    if (expec_id == 0)
    {
        fprintf(stderr,"unable to subscribe to counter: %s", event_name);
        return -1;
    }
    /* That two lines are NOT threadsafe, but its OK because we are "ONCE" */
    local_id = used_expec_counters;
    used_expec_counters++;

    expec_counters[local_id].expec_id = expec_id;
    return local_id;
}


int is_thread_registered()
{
    uint8_t *flag;
    flag = (uint8_t*)pthread_getspecific(pthreadKey);
    if (flag == NULL)
      return 0;
    else
      return 1;
}

uint64_t get_current_value(int32_t ID){
  uint64_t value;
  value=expec_counters[ID].last_value.uint64;
  return value;
}


vt_plugin_cntr_info get_info()
{
    vt_plugin_cntr_info info;
    memset(&info,0,sizeof(vt_plugin_cntr_info));
    info.init                     = init;
    info.is_thread_registered     = is_thread_registered;
    info.add_counter              = add_counter;

    info.vt_plugin_cntr_version   = VT_PLUGIN_CNTR_VERSION;
    info.run_per                  = VT_PLUGIN_CNTR_ONCE;
    info.synch                    = VT_PLUGIN_CNTR_SYNCH;
    info.get_event_info           = get_event_info;
    info.get_current_value        = get_current_value;
    info.finalize                 = fini;
    return info;
}
